###############################################################################
#
# Copyright (c) 2012 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

__author__ = 'Ruslan Spivak <ruslan.spivak@gmail.com>'

import socket
import logging
from logging import handlers

import logsna

LOG_UNIX_SOCKET = '/tmp/_cynic.sock'


class LogUnixSocketHandler(handlers.SocketHandler):
    """Sends pickled log records over a Unix domain socket."""

    def __init__(self, host=LOG_UNIX_SOCKET):
        """host - a path to socket"""
        handlers.SocketHandler.__init__(self, host, None)

    def makeSocket(self, timeout=1):
        """Overriden factory method to create Unix Domain Stream socket."""
        s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        if hasattr(s, 'settimeout'):
            s.settimeout(timeout)
        s.connect(self.host)
        return s


def get_stream_logger(name, level=logging.DEBUG):
    """Logger for children that serve client connections.

    It uses LogUnixSocketHandler to communicate with parent's
    logging server.
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    socket_handler = LogUnixSocketHandler()
    # don't bother with a formatter, since a socket handler sends the event as
    # an unformatted pickle
    logger.addHandler(socket_handler)
    return logger


def get_console_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    # create an instance of the sane formatter
    formatter = logsna.Formatter()

    # add our formatter to the console handler
    ch.setFormatter(formatter)

    # add the console handler to the logger
    logger.addHandler(ch)

    return logger



###############################################################################
#
# Copyright (c) 2012 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

__author__ = 'Ruslan Spivak <ruslan.spivak@gmail.com>'

from http.server import BaseHTTPRequestHandler

from cynic.utils import get_stream_logger  # do not call at module level


class BaseHandler(object):
    """Base handler class for stream sockets."""

    LOGGER_NAME = __name__

    def __init__(self, connection, client_address):
        """
        Args:
            connection - connected socket returned by server's accept
            client_address - tuple containing client_host and client_port
        """
        self.connection = connection
        self.client_address = client_address
        self.logger = get_stream_logger(self.LOGGER_NAME)

    def handle(self):
        pass


class BaseHTTPHandler(BaseHTTPRequestHandler):
    """Base handler class for HTTP protocol."""

    protocol_version = 'HTTP/1.0'

    CONTENT_TYPE = 'text/plain'
    TEMPLATE = ''
    LOGGER_NAME = __name__

    def __init__(self, connection, client_address,
                 datapath=None, content_type=None):
        """
        Args:
            connection - connected socket returned by server's accept
            client_address - tuple containing client_host and client_port
            datapath - file path to a data file that will be sent as
                       as a response body to the client
            content_type - HTTP response Content-Type header value
        """
        self.connection = self.request = connection
        self.client_address = client_address
        self.server = self
        self.setup()

        self.data = open(datapath).read() if datapath is not None else ''
        self.logger = get_stream_logger(self.LOGGER_NAME)
        if content_type is not None:
            self.CONTENT_TYPE = content_type

    def do_GET(self):
        """HTTP GET request handler"""
        body = self.TEMPLATE
        if self.data:
            body = self.data
        self.send_response(200)
        self.send_header('Content-Length', len(body))
        self.send_header('Content-Type', self.CONTENT_TYPE)
        self.end_headers()
        self.wfile.write(body.encode())

    def log_message(self, format, *args):
        """Overridden method from the base class to use our logger."""
        self.logger.info(format % args)



###############################################################################
#
# Copyright (c) 2012 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

__author__ = 'Ruslan Spivak <ruslan.spivak@gmail.com>'

import pickle
import struct
import logging

from cynic.handlers.base import BaseHandler
from cynic.utils import get_console_logger


# Based on LogRecordStreamHandler example from standard logging documentation
class LogRecordHandler(BaseHandler):
    """Handler for a Unix Domain streaming logging request.

    It logs the logging record using whatever logging policy
    is configured locally.
    """

    def handle(self):
        """Handle multiple requests.

        Each request is expected to be a 4-byte length, followed by
        the LogRecord in pickle format.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack('>L', chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))

            # unpickle
            obj = pickle.loads(chunk)

            # log the record
            record = logging.makeLogRecord(obj)
            self.log_record(record)

    def log_record(self, record):
        logger = get_console_logger(record.name)
        logger.handle(record)



###############################################################################
#
# Copyright (c) 2017 Ruslan Spivak, Maxim Muzafarov, Murad Byashimov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

__author__ = 'Maxim Muzafarov <m.muzafarov@gmail.com>'

import os

from setuptools import setup, find_packages


classifiers = """\
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Programming Language :: Python
Topic :: Software Development :: Testing
Operating System :: Unix
"""


def read(*rel_names):
    return open(os.path.join(os.path.dirname(__file__), *rel_names)).read()


setup(
    name='cynic3',
    version='3.0',
    url='https://github.com/m-messiah/cynic3',
    license='MIT',
    description='Test harness to make your system under test cynical',
    author='Maxim Muzafarov',
    author_email='m.muzafarov@gmail.com',
    packages=find_packages(),
    zip_safe=False,
    install_requires=['logsna'],
    entry_points="""\
    [console_scripts]
    cynic = cynic.server:main
    """,
    classifiers=[x for x in classifiers.split('\n') if x],
    long_description=read('README.rst') + '\n\n' + read('CHANGES.rst'),
)



###############################################################################
#
# Copyright (c) 2012 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

__author__ = 'Ruslan Spivak <ruslan.spivak@gmail.com>'

import os
import sys
import errno
import socket
import select
import signal
import optparse
import io
import configparser

from cynic.utils import LOG_UNIX_SOCKET, get_console_logger

READ_ONLY = select.POLLIN
POLL_TIMEOUT = 500  # 0.5 sec
BACKLOG = 5

DEFAULT_CONFIG = """\
############################################################
# HTTP protocol specific                                   #
############################################################

[handler:httphtml]
# sends simple 'hello world!' HTML page over HTTP as a response
# and terminates
class = cynic.handlers.httphtml.HTTPHtmlResponse
#args = ('/tmp/test.html', )
host = 0.0.0.0
port = 2000

[handler:httpjson]
# sends simple 'hello world!' JSON over HTTP as a response
# and terminates
class = cynic.handlers.httpjson.HTTPJsonResponse
#args = ('/tmp/test.json', )
host = 0.0.0.0
port = 2001

[handler:httpnone]
# sends headers, but not the response body and terminates
class = cynic.handlers.httpnone.HTTPNoBodyResponse
host = 0.0.0.0
port = 2002

[handler:httpslow]
# sends one byte of the response every 30 seconds.
# when the data to be sent is exhausted - terminates
class = cynic.handlers.httpslow.HTTPSlowResponse
#args = ('/tmp/test.json', 'application/json', 1)
host = 0.0.0.0
port = 2003


############################################################
# Any TCP socket protocol                                  #
############################################################

[handler:reset]
# accepts a connection, sends an RST packet right away
# and terminates
class = cynic.handlers.reset.RSTResponse
host = 0.0.0.0
port = 2020

[handler:random]
# accepts a connection, sends 7 bytes from the /dev/urandom device
# and terminates
class = cynic.handlers.rnd.RandomDataResponse
host = 0.0.0.0
port = 2021

[handler:noresp]
# accepts a connection, but doesn't send any response back.
# sleeps for 24 hours and exits
class = cynic.handlers.noresp.NoResponse
host = 0.0.0.0
port = 2022

############################################################
# System handlers used internally by the Cynic server      #
############################################################

[handler:unixlog]
# a logging server that accepts connections over Unix socket
# from multiple local processes to output passed log records
class = cynic.handlers.log.LogRecordHandler
host = %(socketpath)s
port = 0
family = unix
""" % dict(socketpath=LOG_UNIX_SOCKET)

# Setup our console logger
logger = get_console_logger('server')


# taken from logging.config
def _resolve(name):
    """Resolve a dotted name to a global object."""
    name = name.split('.')
    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used = used + '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)
    return found


def _reap_children(signum, frame):
    """Collect zombie children."""
    while True:
        try:
            # wait for all children, do not block
            pid, status = os.waitpid(-1, os.WNOHANG)
            if pid == 0:  # no more zombies
                break
        except:
            # Usually this would be OSError exception
            # with 'errno' attribute set to errno.ECHILD
            # which means there are no more children
            break


def _load_config(fname):
    """Return an instance of ConfigParser."""
    config = configparser.ConfigParser()
    if hasattr(fname, 'readline'):
        config.readfp(fname)
    else:
        config.read(fname)

    return config


class HandlerConfig(object):
    def __init__(self, klass, args, host, port, family):
        self.klass = klass
        self.args = args
        self.host = host
        self.port = port
        self.family = family
        self.socket = None  # set up by the server


def _get_handler_configs(config):
    configs = []
    for section in config.sections():
        if not section.startswith('handler:'):
            continue

        clsname = config.get(section, 'class')
        klass = _resolve(clsname)
        host = config.get(section, 'host')
        port = config.getint(section, 'port')
        args = ()
        if config.has_option(section, 'args'):
            args = eval(config.get(section, 'args'), {})
        if config.has_option(section, 'family'):
            family = config.get(section, 'family')
        else:
            family = 'inet'
        hconfig = HandlerConfig(klass, args, host, port, family)
        configs.append(hconfig)

    return configs


class IOLoop(object):
    """Main IO loop.

    Spawns 'crafty' children to handle client requests.
    """
    def __init__(self, handler_configs):
        self.handler_configs = handler_configs
        self.fd2config = {}
        self.poller = None
        self._setup()

    def _get_address(self, hconfig):
        address = {
            'inet': (hconfig.host, hconfig.port),
            'unix': hconfig.host,
        }.get(hconfig.family)
        return address

    def _get_family(self, hconfig):
        family = {
            'inet': socket.AF_INET,
            'unix': socket.AF_UNIX
        }.get(hconfig.family)
        return family

    def _unlink_file(self, hconfig):
        if hconfig.family == 'unix':
            try:
                os.unlink(hconfig.host)
            except:
                pass

    def _setup(self):
        signal.signal(signal.SIGCHLD, _reap_children)

        self.poller = poller = select.poll()

        for config in self.handler_configs:
            server = socket.socket(
                self._get_family(config), socket.SOCK_STREAM)
            server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server.setblocking(0)

            logger.info(
                'Starting %-20r on port %s',
                config.klass.__name__, config.port or config.host
            )

            self.fd2config[server.fileno()] = config

            # this is only for Unix socket
            self._unlink_file(config)

            server.bind(self._get_address(config))

            server.listen(BACKLOG)

            # save the listen socket for further use
            config.socket = server

            poller.register(server, READ_ONLY)

    def run(self):
        try:
            child_pids = []
            while True:
                try:
                    events = self.poller.poll(POLL_TIMEOUT)
                except select.error as e:
                    code, msg = e.args
                    if code == errno.EINTR:
                        continue
                    else:
                        raise

                for fd, flag in events:
                    # we're interested only in READ events
                    if not flag & READ_ONLY:
                        continue

                    # retrieve the actual socket and handler class
                    # from its file descriptor
                    handler_config = self.fd2config[fd]

                    # socket is ready to accept a connection
                    socket = handler_config.socket
                    try:
                        conn, client_address = socket.accept()
                    except IOError as e:
                        code, msg = e.args
                        if code == errno.EINTR:
                            continue
                        else:
                            raise

                    # spawn a child that will handle the request (connection)
                    pid = os.fork()
                    if pid == 0:  # child
                        for config in self.fd2config.values():
                            config.socket.close()
                        # run a handler
                        klass = handler_config.klass
                        handler = klass(
                            conn, client_address, *handler_config.args)
                        try:
                            handler.handle()
                        except KeyboardInterrupt:
                            pass
                        except:
                            log = get_console_logger(klass.__name__)
                            log.exception('Exception when handling a request')
                        # off we go
                        os._exit(0)
                    else:
                        # this is parent
                        child_pids.append(pid)
                        # close unused connected socket
                        conn.close()
        except KeyboardInterrupt:
            for pid in child_pids:
                try:
                    os.kill(pid, signal.SIGTERM)
                except:
                    pass


def main():
    parser = optparse.OptionParser(
        description='Test harness to make your system under test cynical ')
    parser.add_option(
        '-c', '--config', dest='config_path',
        help=(
            'Path to an INI configuration file. If no configuration '
            'file is provided then default configuration is applied. '
            'To see the default configuration use -d option described below.'
        )
    )
    parser.add_option(
        '-d', '--dump', dest='dump', action='store_true',
        default=False, help='Dump default configuration to STDOUT'
    )

    options, args = parser.parse_args()

    if options.dump:
        print(DEFAULT_CONFIG)
        sys.exit(0)

    path = options.config_path
    if path is None:
        path = io.StringIO(DEFAULT_CONFIG)

    config = _load_config(path)
    handlers = _get_handler_configs(config)
    ioloop = IOLoop(handlers)
    ioloop.run()
